package gkv

import (
	"bufio"
	"fmt"
	"io"
	"reflect"
	"strings"
)

// a k-v is map to a two level map[string](map[string]interface{})
// start with # and empty line will be ingore
// select will like [%s], all space around &s with be remove
// k-v like %s = %s , sapce befor and after %s will be remove
// at list on slecet need

type Selection map[string]interface{}

type Config map[string](*Selection)

func (c *Config) Get(sel string, key string) interface{} {
	if sel, ok := (*c)[sel]; ok {
		if v, ok := (*sel)[key]; ok {
			return v
		}
		return nil
	}
	return nil
}

func (c *Config) Dump() {
	for k, v := range *c {
		fmt.Printf("[%s]\n", k)
		for sk, sv := range *v {
			fmt.Printf("%s = %s\n", sk, sv)
		}
	}
}

func isEmpty(s uint8) bool {
	return s == ' '
}

func removeAroundSpace(str string) string {
	i := 0
	for i = 0; i < len(str); i++ {
		if !isEmpty(str[i]) {
			break
		}
	}

	str = str[i:len(str)]

	for i = len(str) - 1; i > 0; i-- {
		if !isEmpty(str[i]) {
			break
		}
	}
	str = str[0 : i+1]
	return str
}

const (
	TypeComment   = 0
	TypeSelection = 1
	TypeKeyValue  = 2
	TypeEmpty     = 3
	TypeError     = 4
)

type Liner struct {
	Type  int
	Key   string
	Value string
}

func parseKeyValue(l string) (Liner, error) {
	equalIndex := strings.Index(l, "=")
	if -1 == equalIndex {
		return Liner{Type: TypeError}, fmt.Errorf("parse Selection Failed: %s", l)
	}
	key := removeAroundSpace(l[0:equalIndex])
	val := removeAroundSpace(l[equalIndex+1 : len(l)])
	if 0 == len(key) {
		return Liner{Type: TypeError}, fmt.Errorf("parse Selection Failed: %s", l)
	}
	return Liner{Type: TypeKeyValue, Key: key, Value: val}, nil
}

func parseSelection(l string) (Liner, error) {
	if !('[' == l[0] && ']' == l[len(l)-1]) {
		return Liner{Type: TypeError}, fmt.Errorf("parse Selection Failed: %s", l)
	}
	l = l[1 : len(l)-1]
	sel := removeAroundSpace(l)
	if 0 == len(sel) {
		return Liner{Type: TypeError}, fmt.Errorf("parse Selection Failed, empty selection name")
	}
	return Liner{Type: TypeSelection, Key: "Selection", Value: sel}, nil
}

func parseLine(l string) (Liner, error) {
	l = removeAroundSpace(l)
	if 0 == len(l) {
		return Liner{Type: TypeEmpty}, nil
	}

	switch l[0] {
	case '[':
		return parseSelection(l)
	case '#':
		return Liner{Type: TypeComment}, nil
	}
	return parseKeyValue(l)
}

func ParseInto(v interface{}, r io.Reader) error {
	lr := bufio.NewReader(r)
	var curSel *Selection
	var curSelName string
	cfg := Config{}
	for {
		l, _, err := lr.ReadLine()
		if nil != err {
			if io.EOF == err {
				vcfg := reflect.ValueOf(v)
				gkvc := vcfg.Elem().FieldByName("Config")
				if gkvc.IsValid() {
					gkvc.Set(reflect.ValueOf(cfg))
				}
				return nil
			}

			return fmt.Errorf("parse %v Failed %v", l, err)
		}
		liner, err := parseLine(string(l))
		if nil != err {
			return fmt.Errorf("parse %v Failed %v", l, err)
		}
		switch liner.Type {
		case TypeComment:
			continue
		case TypeEmpty:
			continue
		case TypeError:
			return fmt.Errorf("parse %v Failed", l)
		case TypeSelection:
			curSel = &Selection{}
			curSelName = liner.Value
			cfg[liner.Value] = curSel
			continue
		case TypeKeyValue:
			if nil == curSel {
				return fmt.Errorf("parse %v Failed, Empty Selection", l)
			}
			set(v, curSelName, "", liner.Key, false, liner.Value)
			(*curSel)[liner.Key] = liner.Value
			continue
		}
	}
	return nil
}
